
# importing sys
import sys
  
# adding Folder_2 to the system path
sys.path.insert(0, '../')
import UQLibrary as uq
#import mpi4py.MPI as MPI

def main():
    # #Set seed for reporducibility
    # #np.random.seed(10)
    
    # # Get model and options object from Example set
    # [model, options] = uqExamples.GetExample('aluminum rod (uniform)')
    #
    # # [model, options] = uqExamples.GetExample('linear product')
    #
    # [model, options] = uqExamples.GetExample('aluminum rod (saltelli normal)')
    #options.plot.nPoints=options.gsa.nSamp
    #f
    # [model, options] = uqExamples.GetExample('aluminum rod (normal)')
    #
    example = 'Helmholtz (double unidentifiable)'
    print("Loading Example:" + str(example) + " function.")
    # [model, options] = uq.examples.GetExample('ishigami (normal)')
    # options.lsa.run_param_subset = False
    # options.display = True
    # options.save = True
    [model, options] = uq.examples.GetExample(example)
    options.lsa.run_lsa = False
    options.lsa.run_param_subset = True
    options.gsa.run = False
    options.display = True
    options.save = True
    #options.plot = True
    #
    # # Run UQ package
    # (baseSobol,totalSobol)= uq.TestAccuracy(model, options, np.arange(start=10000, stop=200000, step=5000))
    results = uq.run_uq(model, options, logging = 3)
    #
    # plt.plot(results.gsa.sampD[:450,0], results.gsa.sampD[:450,1],'rs')
    # plt.plot(results.gsa.sampD[options.gsa.nSamp:options.gsa.nSamp+450,0], results.gsa.sampD[options.gsa.nSamp:options.gsa.nSamp+450,1],'bo')
    # plt.show()
if __name__ == '__main__':
    main()





# UQtoolbox_examples
# Module of example problems setup for running in UQtoolbox
#   Each example outputs a model and options object with all required fields

import UQLibrary as uq
import numpy as np
import math as math
import scipy.integrate as integrate

def GetExample(example, **kwargs):
    # Master function for selecting an example using a corresponding string
    # Inputs: example- string that corresponds to the desired model
    # Outputs: model and options objects corresponding to the desired model

    # Initialize options object
    options = uq.Options()
    # Select Example model
    if example.lower() == 'linear':
        #baseEvalPoints = np.array([0, .5, 1, 2])  # Requires 1xnQOIs indexing
        baseEvalPoints= np.array([2])
        model = uq.Model(eval_fcn=lambda params: linear_function(baseEvalPoints, params),
                         base_poi=np.array([1, 1]),
                         cov=np.array([[1, 0], [0, 1]]),
                         dist_type='uniform',
                         dist_param=np.array([[0], [1]])*np.array([1, 1])
                         #dist_param=np.array([[.9999999999], [1.0000000001]])*np.array([1, 1])
                         )
    
    elif example.lower() == 'quadratic':
        baseEvalPoints = np.array([0, .5, 1, 2])  # Currently requires 1xn or nx1 ordering
        model = uq.Model(eval_fcn=lambda params: quadratic_function(baseEvalPoints, params),
                         base_poi=np.array([1, 1, 1]),
                         cov=np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]]))
        options.gsa.n_samp_sobol=100               # Keep normal sampling but reduce sample size to 100
    
    elif example.lower() == 'helmholtz (identifiable)':
        baseEvalPoints = np.linspace(0,1,100)
        model = uq.Model(eval_fcn=lambda params: HelmholtzEnergy(baseEvalPoints, params),
                         base_poi=np.array([-392.66, 770.1, 57.61]),
                         cov=np.array([[0.0990, - 0.4078, 0.4021],  # Covaraince matrix calculated by DRAMs
                                       [-0.4078, 2.0952, -2.4078],  # at baseParams and basEvalPoints
                                       [0.4021, -2.4078, 3.0493]]) * (10 ** 3),
                         name_poi = np.array(["alpha1", "alpha11", "alpha111"]),
                         dist_type='uniform')  # Use uniform sampling of +-20% nominal value
        model.dist_param = np.array([[.8, .8, .8], [1.2, 1.2, 1.2]]) * model.base_poi
        options.gsa.n_samp_sobol=10000              # Keep normal sampling but reduce sample size to 100
    
    elif example.lower() == 'helmholtz (unidentifiable)':
        baseEvalPoints = np.linspace(0,.05,100)
        model = uq.Model(eval_fcn=lambda params: HelmholtzEnergy(baseEvalPoints, params),
                         base_poi=np.array([-392.66, 770.1, 57.61]),
                         cov=np.array([[0.0990, - 0.4078, 0.4021],  # Covaraince matrix calculated by DRAMs
                                       [-0.4078, 2.0952, -2.4078],  # at baseParams and basEvalPoints
                                       [0.4021, -2.4078, 3.0493]]) * (10 ** 3),
                         name_poi = np.array(["alpha1", "alpha11", "alpha111"]),
                         dist_type='uniform')  # Use uniform sampling of +-20% nominal value
        model.dist_param = np.array([[.8, .8, .8], [1.2, 1.2, 1.2]]) * model.base_poi
        options.gsa.n_samp_sobol=10000              # Keep normal sampling but reduce sample size to 100
        options.lsa.pss_rel_tol = 1e-6
    
    elif example.lower() == 'helmholtz (double unidentifiable)':
        baseEvalPoints = np.linspace(0,.05,100)
        model = uq.Model(eval_fcn=lambda params: HelmholtzEnergy(baseEvalPoints, params[0:3]),
                         base_poi=np.array([-392.66, 770.1, 57.61, 1]),
                         cov=np.array([[0.0990, - 0.4078, 0.4021, 0],  # Covaraince matrix calculated by DRAMs
                                       [-0.4078, 2.0952, -2.4078, 0],  # at baseParams and basEvalPoints
                                       [0.4021, -2.4078, 3.0493, 0],
                                       [0, 0, 0, 1]]) * (10 ** 3),
                         name_poi = np.array(["alpha1", "alpha11", "alpha111", "fake parameter"]),
                         dist_type='uniform')  # Use uniform sampling of +-20% nominal value
        model.dist_param = np.array([[.8, .8, .8, .8], [1.2, 1.2, 1.2, 1.2]]) * model.base_poi
        options.gsa.n_samp_sobol=10000              # Keep normal sampling but reduce sample size to 100
        options.lsa.pss_rel_tol = 1e-6
        
    elif example.lower() == 'integrated helmholtz':
        baseEvalPoints=np.arange(0,.8,.06)
        model = uq.Model(eval_fcn=lambda params: IntegratedHelmholtzEnergy(baseEvalPoints, params),
                         base_poi=np.array([-389.4, 761.3, 61.5]),
                         cov=np.array([[0.0990, - 0.4078, 0.4021],  # Covaraince matrix calculated by DRAMs
                                       [-0.4078, 2.0952, -2.4078],  # at baseParams and basEvalPoints
                                       [0.4021, -2.4078, 3.0493]]) * (10 ** 3),
                         name_poi=np.array(["alpha_1", "alpha11", "alpha111"]),
                         name_qoi=np.array(["x=.8", "x=.80001"]),
                         dist_type="uniform")  # Use uniform sampling of +-20% nominal value
        model.dist_param = np.array([[.8, .8, .8], [1.2, 1.2, 1.2]]) * model.base_poi
        #model.dist_param = np.array([[.999999, .999999, .999999], [1.000001, 1.000001, 1.000001]]) * model.base_poi
    
    elif example.lower() == 'linear product':  # Linear product example taken from Homma1996
        model = uq.Model(eval_fcn=LinearProd,
                         base_poi=np.array([.5, .5, .5, .5, .5]),
                         dist_type="uniform",
                         dist_param=np.array([[0, 0, 0, 0, 0], [1, 1, 1, 1, 1]]))
        
    elif example.lower() == 'ishigami (uniform)':
        model = uq.Model(eval_fcn=Ishigami,
                         base_poi=np.array([0, 0, 0]),
                         dist_type="saltelli uniform",
                         dist_param=np.array([[-math.pi, -math.pi, -math.pi], [math.pi, math.pi, math.pi]]))
        options.lsa.deriv_method = 'finite' 
        options.lsa.xDelta = 10**(-6)
        options.gsa.n_samp_sobol = 500000          # Use default number of samples
        
    elif example.lower() == 'ishigami (normal)':
        model = uq.Model(eval_fcn=Ishigami,
                         base_poi=np.array([0, 0, 0]),
                         dist_type="saltelli normal",
                         dist_param=np.array([[0, 0, 0], [(2*math.pi)**2/12, (2*math.pi)**2/12, (2*math.pi)**2/12]]))
        options.lsa.deriv_method = 'finite' 
        options.lsa.xDelta = 10**(-6)
        options.gsa.n_samp_sobol= 500000 
        
    elif example.lower() == 'trial function':
        model = uq.Model(eval_fcn=TrialFunction,
                         base_poi=np.array([1, 1, 1]),
                         dist_type="uniform",
                         dist_param=np.array([[1, 1, 1], [1000, 100, 10]])
                         )
    #---------------------------------Portfolio Models------------------------------------------------------------
    # See p.353 of Smith, 2015, Uncertainty Quantification
    elif example.lower() == 'portfolio (normal)':
        model=uq.Model(eval_fcn=lambda params: Portfolio(params, np.array([2, 1])),
                       base_poi=np.array([0, 0]),
                       dist_type="normal",
                       dist_param=np.array([[0, 0], [1, 9]]))
        options.gsa.n_samp_sobol = 100000
        
    elif example.lower() == 'portfolio (uniform)':
        model=uq.Model(eval_fcn=lambda params: Portfolio(params, np.array([2, 1])),
                       base_poi=np.array([0, 0]),
                       dist_type="uniform",
                       dist_param=np.array([[-np.sqrt(12)/2, -3*np.sqrt(3)], [np.sqrt(12)/2, 3*np.sqrt(3)]]))
        options.path = '..\\Figures\\Portfolio(Uniform)'
        options.gsa.n_samp_sobol = 2**12
    #--------------------------------Heated Rod Models------------------------------------------------------------
    elif example.lower() == 'aluminum rod (uniform)':
        model = uq.Model(eval_fcn=lambda params: HeatRod(params, np.array([55])),
                         base_poi=np.array([-18.4, .00191]),
                         dist_type="uniform",
                         dist_param=np.array([[-18.4-(.1450*np.sqrt(3)), .00191-(1.4482*(10**(-5))*np.sqrt(3))],\
                                             [-18.4+(.1450*np.sqrt(3)), .00191+(1.4482*(10**(-5))*np.sqrt(3))]]),
                         name_poi=np.array(['Phi', 'h']),
                         name_qoi=np.array(['T(x=55)']))
        options.path = '..\\Figures\\AluminumRod(Uniform, x=55)'
        options.gsa.n_samp_sobol = 500000
        
    elif example.lower() == 'aluminum rod (normal)':
        model = uq.Model(eval_fcn=lambda params: HeatRod(params, np.array([15, 25, 35, 45, 55])),
                         base_poi=np.array([-18.4, .00191]),
                         dist_type='normal',
                         dist_param=np.array([[-18.4, .00191], [.1450**2, (1.4482*10**(-5))**2]]),
                         name_poi=np.array(['Phi', 'h']),
                         name_qoi=np.array(["x=15", "x=25", "x=35", "x=45", "x=55"])
                         #name_qoi=np.array(['T(x=15)','T(x=55)'])
                         )
        options.path = '..\\Figures\\AluminumRod(Normal)'
        options.gsa.n_samp_sobol = 100000000
        
    elif example.lower() == 'aluminum rod (saltelli normal)':
        model = uq.Model(eval_fcn=lambda params: HeatRod(params, np.array([55])),
                         base_poi=np.array([-18.4, .00191]),
                         dist_type="saltelli normal",
                         dist_param=np.array([[-18.4, .00191], [.1450**2, (1.4482*10**(-5))**2]]),
                         name_poi=np.array(['Phi', 'h']),
                         name_qoi=np.array(['T(x=55)']))
        options.path = '..\\Figures\\AluminumRod(saltelli_normal, x=55)'
        options.gsa.n_samp_sobol = 200000
    #------------------------------------SIR Models----------------------------------------------------------------
        
    elif example.lower() == 'sir infected':
        model = uq.Model(eval_fcn = lambda params: SolveSIRinfected(params, np.array([960, 40, 0]), np.array([0, 1, 3, 5, 6])),
                         base_poi=np.array([8, 1.5]),
                         name_poi=np.array(['beta', 'gamma']),
                         dist_type='uniform',
                         dist_param=np.array([[0, 0], [1, 1]])
                         )
        options.lsa.deriv_method='finite'
        options.lsa.xDelta=.00001
    elif example.lower() == 'sir enedmic':
            model = uq.Model(eval_fcn = lambda params: SIR_endemic_integrated(params, np.array([900, 100, 0]), 20),
                             base_poi=np.array([8, 1.5]),
                             name_poi=np.array(['gamma', 'k', 'r', 'delta']),
                             dist_type='uniform',
                             dist_param=np.array([[0, 0], [1, 1]])
                             )
            options.lsa.method='finite'
            options.lsa.xDelta=.00001
        
    elif example.lower() == 'sobol test function':
        model = uq.Model(eval_fcn= lambda params: SobolTestFunction(params,np.array([78, 12, .5, 2, 97, 33])),
                         base_poi= np.array([.5, .5, .5, .5, .5, .5]),
                         dist_type='uniform',
                         dist_param=np.array([[0,0,0,0,0,0], [1,1,1,1,1,1]]))
    else:
        raise Exception("Unrecognized Example Type")

    # Apply optional inputs
    if 'basePOI' in kwargs:  # Change base parameter values to input
        model.base_poi = kwargs['basePOI']
    if 'evalPoints' in kwargs:  # Determine eval points
        model.evalPoints = kwargs['evalPoints']

    return model, options


def linear_function(x, params):
    if params.ndim == 1:
        return params[0] + (x * params[1])
    if params.ndim == 2:
        return params[:, 0] + (params[:, 1] * x)


def quadratic_function(x, params):
    if params.ndim == 1:
        return params[0] + (x * params[1]) + ((x ** 2) * params[2])
    if params.ndim == 2:
        return np.outer(params[:, 0], np.ones(x.shape)) + np.outer(params[:, 1], x) + np.outer(params[:, 2], x**2)


def HelmholtzEnergy(x, params):
    if params.ndim == 1:
        return params[0] * (x ** 2) + params[1] * (x ** 4) + params[2] * (x ** 6)
    elif params.ndim == 2:
        return params[:, 0] * (x ** 2) + params[:, 1] * (x ** 4) + params[:, 2] * (x ** 6)


def IntegratedHelmholtzEnergy(x, params):
    if params.ndim == 1:
        return params[0] * (x ** 3) / 3 + params[1] * (x ** 5) / 5 + params[2] * (x ** 7) / 7
    elif params.ndim == 2:
        return np.outer(params[:, 0], (x ** 3)) / 3 + np.outer(params[:, 1], (x ** 5)) / 5 + np.outer(params[:, 2], (x ** 7)) / 7



def LinearProd(params):
    if params.ndim == 1:
        return np.array([np.prod(2 * params + 1) / (2 ** (len(np.transpose(params))))])
    elif params.ndim == 2:
        return np.prod(2 * params + 1, axis=1) / (2 ** (len(np.transpose(params)) + 1))
def Ishigami(params):
    if params.ndim == 1:
        return np.array([np.sin(params[0])+7*np.sin(params[1])**2+.1*(params[2]**4)*np.sin(params[0])])
    elif params.ndim == 2:
        return np.sin(params[:, [0]])+7*np.sin(params[:, [1]])**2+.1*(params[:, [2]]**4)*np.sin(params[:, [0]])
def TrialFunction(params):
    if params.ndim == 1:
        return np.array([params[0]+params[1]*(params[2]**2)])
    elif params.ndim == 2:
        return params[:, [0]]+params[:, [1]]*(params[:, [2]]**2)
def Portfolio(params,c):
    if params.ndim == 1:
        return np.array([c[0]*params[0]+c[1]*params[1]])
    elif params.ndim == 2:
        return c[0]*params[:, 0]+c[1]*params[:, 1]
def HeatRod(params,x):
    #Set or load paramters
    Tamb=21.19
    a=.95
    b=.95
    L=70
    k=2.37
    if params.ndim==1:
        Phi=params[0]
        h=params[1]
    elif params.ndim==2:
        Phi=params[:,0]
        h=params[:,1]
    #Compute intermediates
    gamma=np.sqrt(2*(a+b)*h/(a*b*k))
    c1=-Phi/(k*gamma)*(np.exp(gamma*L)*(h+k*gamma)/(np.exp(-gamma*L)*(h-k*gamma)+np.exp(gamma*L)*(h+k*gamma)))
    c2=Phi/(k*gamma)+c1
    #Compute temperature
    gx = np.outer(gamma, x)
    if (gx.shape[0] != 1 and gx.shape[1] == 1) or (gx.shape[1] != 1 and gx.shape[0] == 1):
        gx = gx.squeeze()
    if gx.ndim == 2 and c1.ndim == 1:
        c1 = c1[:, np.newaxis]
        c2 = c2[:, np.newaxis]
    T = c1*np.exp(-gx)+c2*np.exp(gx)+Tamb
    return T
def SolveSIRinfected(params,y0,tEval):
    if params.ndim==1:
        sol=integrate.solve_ivp(lambda t,y: SIRdydt(params,t,y), np.array([0, np.max(tEval)]),y0,t_eval=tEval)
        infected=sol.y[1,:]
    else:
        infected=np.empty((params.shape[0],tEval.size))
        for i in np.arange(0,params.shape[0]):
            sol = integrate.solve_ivp(lambda t, y: SIRdydt(params[i,], t, y), np.array([0, np.max(tEval)]), y0,
                                      t_eval=tEval)
            infected[i,:]=sol.y[1,:]
    return infected


def SIR_endemic_integrated(params,y0,tEval):
    if params.ndim==1:
        sol=integrate.solve_ivp(lambda t,y: SIRdydt_endemic(params,t,y), np.array([0, np.max(tEval)]),y0,t_eval=tEval)
        recovered=sol.y[2,:]
        tdiff = sol.t[1:] - sol.t[:-1]
        rhr = np.sum(recovered[1:]*tdiff)
        lhr = np.sum(recovered[:-1]*tdiff)
        total_infected = (rhr+lhr)/2
    else:
        infected=np.empty((params.shape[0]))
        for i in np.arange(0,params.shape[0]):
            sol = integrate.solve_ivp(lambda t, y: SIRdydt_endemic(params[i,], t, y), np.array([0, np.max(tEval)]), y0,
                                      t_eval=tEval)
            recovered=sol.y[2,:]
            tdiff = sol.t[1:] - sol.t[:-1]
            rhr = np.sum(recovered[1:]*tdiff)
            lhr = np.sum(recovered[:-1]*tdiff)
            total_infected[i,:] = (rhr+lhr)/2
    return infected

def SIRdydt_endemic(params,t,y):
    dydt=np.empty(3)
    dydt[0] = params[3]*(y[1]+y[2])-params[0]*params[1]*y[0]*y[1]
    dydt[1] = params[0]*params[1]*y[0]*y[1] - (params[3]+params[2])*y[1]
    dydt[2] = params[2]*y[1]-params[3]*y[2]
    return dydt

def SIRdydt(params,t,y):
    dydt=np.empty(3)
    dydt[0]=-params[0]*y[1]/(np.sum(y))*y[0]
    dydt[1]=params[0]*y[1]/(np.sum(y))*y[0]-params[1]*y[1]
    dydt[2]=params[1]*y[1]
    return dydt

def SobolTestFunction(theta,a):
    if theta.ndim==1:
        return np.array([np.prod((np.abs(4*theta-2)+a)/(1+a))])
    else:
        return np.prod((np.abs(4*theta-2)+a)/(1+a), axis=1)


# -*- coding: utf-8 -*-
"""
Created on Tue Jan 18 14:03:35 2022

@author: USER
"""

#3rd party Modules
import numpy as np
#import sys
import warnings
import mpi4py.MPI as MPI
from .sampling import parallel_eval


class GsaOptions:
    __slots__=["run", "run_sobol", "run_morris", "n_samp_sobol", "n_samp_morris",\
               "l_morris"]
    def __init__(self, run = True, run_sobol=True, run_morris=True, n_samp_sobol=100000, \
                 n_samp_morris=4, l_morris=3):
        self.run = run
        if self.run == False:
            self.run_sobol = False
            self.run_morris = False
        else:
            self.run_sobol=run_sobol                            #Whether to run Sobol (True or False)
            self.run_morris=run_morris                          #Whether to run Morris (True or False)
        self.n_samp_sobol = n_samp_sobol                      #Number of samples to be generated for GSA
        self.n_samp_morris = n_samp_morris
        if callable(l_morris):
            self.l_morris = l_morris
        elif type(l_morris) == int or type(l_morris) == float:
            self.l_morris = l_morris
        else:
            warnings.warn("Unrecongized l_morris type: " + str(type(l_morris)))
        pass

class GsaResults:
    #
    def __init__(self,sobol_base=np.nan, sobol_tot=np.nan, f_a=np.nan, f_b=np.nan, f_d=np.nan, f_ab=np.nan, \
                 samp_d=np.nan, morris_std=np.nan, morris_mean_abs=np.nan, morris_mean=np.nan):
        self.sobol_base=sobol_base
        self.sobol_tot=sobol_tot
        self.f_a=f_a
        self.f_b=f_b
        self.f_d=f_d
        self.f_ab=f_ab
        self.samp_d=samp_d
        self.morris_mean_abs=morris_mean_abs
        self.morris_mean = morris_mean
        self.morris_std=morris_std
    pass



##--------------------------------------GSA-----------------------------------------------------
def run_gsa(model, gsa_options, logging = False):
    """Implements global sensitivity analysis using Morris or Sobol analysis.
    
    Parameters
    ----------
    model : Model
        Contaings simulation information.
    gsa_options : GSAOptions
        Contains run settings
        
    Returns
    -------
    GsaResults 
        Holds all run results
    """
    #GSA implements the following local sensitivity analysis methods on "model" object
        # 1) Gets sampling distribution (used only for internal calculations)
        # 2) Calculates Sobol Indices
        # 3) Performs Morris Screenings (not yet implemented)
        # 4) Produces histogram plots for QOI values (not yet implemented)
    # Required Inputs: Object of class "model" and object of class "options"
    # Outputs: Object of class gsa with fisher and sobol elements
    
    #Load mpi details to keep track of thread number
    mpi_comm = MPI.COMM_WORLD
    mpi_rank = mpi_comm.Get_rank()
    mpi_size = mpi_comm.Get_size()
    
    # Initialize gsa_results in all threads
    gsa_results = GsaResults()
    
    #Morris Screening
    if gsa_options.run_morris:
        #Set non-biased perturbation distance for even l
        #Source: Smith, R. 2011. Uncertainty Quanitification. p.333
        pert_distance = gsa_options.l_morris/ (2*(gsa_options.l_morris-1))
        
        #Create parameter sample only on thread 0 since it need not be parallelized
        # initialize memory location on all threads
        morris_samp = np.zeros((gsa_options.n_samp_morris*(model.n_poi+1), model.n_poi),dtype = float)
        if logging > 1:
            print("initialized morris_samp of size: " + str(morris_samp.shape))
        if mpi_rank == 0:
            if logging:
                print("Generating Morris Sample")
            morris_samp = get_morris_poi_sample(model.sample_fcn, gsa_options.n_samp_morris,\
                                                model.n_poi, pert_distance)
        mpi_comm.Bcast([morris_samp,MPI.DOUBLE], root = 0)
                
            
        morris_mean_abs, morris_mean, morris_std = calculate_morris(\
                                             model.eval_fcn, morris_samp, \
                                             pert_distance, logging = logging)
        gsa_results.morris_mean_abs=morris_mean_abs
        gsa_results.morris_mean = morris_mean
        gsa_results.morris_std=morris_std

    #Sobol Analysis Un parallelized for now
    if gsa_options.run_sobol and mpi_rank == 0:
        if logging:
            print("Generating Sobol Sample")
        #Make Distribution Samples and Calculate model results
        [f_a, f_b, f_ab, f_d, samp_d] = get_sobol_sample(model, gsa_options)
        #Calculate Sobol Indices
        if logging:
            print("Calculating Sobol Sample")
        [sobol_base, sobol_tot]=calculate_sobol(f_a, f_b, f_ab, f_d)
        gsa_results.f_d=f_d
        gsa_results.f_a=f_a
        gsa_results.f_b=f_b
        gsa_results.f_ab=f_ab
        gsa_results.samp_d=samp_d
        gsa_results.sobol_base=sobol_base
        gsa_results.sobol_tot=sobol_tot
        #------------broadcast gsa results to other threads--------------------
        
    return gsa_results


###----------------------------------------------------------------------------------------------
###-------------------------------------Support Functions----------------------------------------
###----------------------------------------------------------------------------------------------


def get_sobol_sample(model,gsa_options):
    """Constructs and evaluates sobol samples using predefined sampling distributions.
        Currently only function for uniform or saltelli normal
    
    Parameters
    ----------
    model : Model
        Contaings simulation information.
    gsa_options : GSAOptions
        Contains run settings
        
    Returns
    -------
    np.ndarray
        n_samp_sobol x n_qoi array of evaluations of Sobol sample part a
    np.ndarray
        n_samp_sobol x n_qoi array of evaluations of Sobol sample part b
    np.ndarray
        n_samp_sobol x n_qoi array x n_poi array of evaluations of mixed Sobol sample ab
    np.ndarray
        2*n_samp_sobol x n_qoi array of concatenated evaluations of part a and b
    np.ndarray
        2*n_samp_sobol x n_poi array of concatenated POI samples of part a and b
    """
    n_samp_sobol = gsa_options.n_samp_sobol
    # Make 2 POI sample matrices with n_samp_sobol samples each
    # if np.all(model.dist_type!=np.array(["satelli normal", "satelli uniform"])):
    #           warnings.warn("Non-satelli sampling algorithm used for Sobol analysis."\
    #                         + " Suggested distribution types are satelli normal "+\
    #                             "and satelli uniform.")
    sample_compact = model.sample_fcn(2*n_samp_sobol)
    f_compact = model.eval_fcn(sample_compact)
    # Seperate sample into a and b for algorithm
    samp_a = sample_compact[:n_samp_sobol]
    samp_b = sample_compact[n_samp_sobol:]
    f_a = f_compact[:n_samp_sobol]
    f_b = f_compact[n_samp_sobol:] # n_samp_sobol x nQOI out matrix from B
    # Stack the output matrices into a single matrix
    f_d = np.concatenate((f_a.copy(), f_b.copy()), axis=0)

    # Initialize combined QOI sample matrices
    if model.n_qoi == 1:
        f_ab = np.empty([n_samp_sobol, model.n_poi])
    else:
        f_ab = np.empty([n_samp_sobol, model.n_poi, model.n_qoi])
    for i_param in range(0, model.n_poi):
        # Define sampC to be A with the ith parameter in B
        samp_ab = samp_a.copy()
        samp_ab[:, i_param] = samp_b[:, i_param].copy()
        if model.n_qoi == 1:
            f_ab[:, i_param] = model.eval_fcn(samp_ab).squeeze()
        else:
            f_ab[:, i_param, :] = model.eval_fcn(samp_ab)  # n_samp_sobol x nPOI x nQOI tensor
        del samp_ab
    return f_a, f_b, f_ab, f_d, sample_compact

def calculate_sobol(f_a, f_b, f_ab, f_d):
    """Calculates 1st order and total sobol indices using Saltelli approximation formula.
    
    Parameters
    ----------
    f_a : np.ndarray
        n_samp_sobol x n_qoi array of evaluations of Sobol sample part a
    f_b : np.ndarray
        n_samp_sobol x n_qoi array of evaluations of Sobol sample part b
    f_ab : np.ndarray
        n_samp_sobol x n_qoi array x n_poi array of evaluations of mixed Sobol sample ab
    f_d : np.ndarray
        2*n_samp_sobol x n_qoi array of concatenated evaluations of part a and b
    
        
    Returns
    -------
    np.ndarray
        n_qoi x n_poi array of 1st order Sobol indices
    np.ndarray
        n_qoi x n_poi array of total Sobol indices
    """
    #Calculates calculates sobol indices using satelli approximation method
    #Inputs: model object (with eval_fcn, sample, and nParams)
    #        sobolOptions object
    #Determing number of samples, QOIs, and POIs based on inputs
    if f_ab.ndim==1:
        n_qoi=1
        n_poi=1
    elif f_ab.ndim==2:
        n_qoi=1
        n_poi=f_ab.shape[1]
    elif f_ab.ndim==3:
        n_poi=f_ab.shape[1]
        n_qoi=f_ab.shape[2]
    else:
        raise(Exception('f_ab has greater than 3 dimensions, make sure f_ab is' \
                        'the squeezed form of n_samp_sobol x nPOI x nQOI'))
    #QOI variance
    fDvar=np.var(f_d, axis=0)

    sobol_base=np.empty((n_qoi, n_poi))
    sobol_tot=np.empty((n_qoi, n_poi))
    if n_qoi==1:
        #Calculate 1st order parameter effects
        sobol_base=np.mean(f_b*(f_ab-f_a), axis=0)/(fDvar)

        #Caclulate 2nd order parameter effects
        sobol_tot=np.mean((f_a-f_ab)**2, axis=0)/(2*fDvar)

    else:
        for iQOI in range(0,n_qoi):
            #Calculate 1st order parameter effects
            sobol_base[iQOI,:]=np.mean(f_b[:,[iQOI]]*(f_ab[:,:,iQOI]-f_a[:,[iQOI]]),axis=0)/fDvar[iQOI]
            #Caclulate 2nd order parameter effects
            sobol_tot[iQOI,:]= np.mean((f_a[:,[iQOI]]-f_ab[:,:,iQOI])**2,axis=0)/(2*fDvar[iQOI])


    return sobol_base, sobol_tot
#==============================================================================
#----------------------------------Morris Sampling-----------------------------
#==============================================================================


##--------------------------------calculate_morris-----------------------------
def calculate_morris(eval_fcn, morris_samp, pert_distance, logging = False):
    """Calculates morris samples using information from Model and GsaOptions objects.
    
    Parameters
    ----------
    model : Model
        Contaings simulation information.
    gsa_options : GSAOptions
        Contains run settings
        
    Returns
    -------
    np.ndarray
        n_qoi x n_poi array of morris sensitivity mean indices
    np.ndarray
        n_qoi x n_poi array of morris sensitivity variance indices
    """
    #Evaluate Sample
    #Load mpi details to keep track of thread number
    mpi_comm = MPI.COMM_WORLD
    mpi_rank = mpi_comm.Get_rank()
    mpi_size = mpi_comm.Get_size()
    if logging and mpi_rank == 0:
        print("Evaulating Morris Sample")
        
    if mpi_size == 1:
        f_eval_compact = eval_fcn(morris_samp)
    else:
        f_eval_compact = parallel_eval(eval_fcn, morris_samp, logging = logging)
        
    #Make sure all threads finish collecting f_eval_compact before continuing
    mpi_comm.Barrier()
    if logging > 1 and mpi_rank == 0:
        print("f_eval_compact: " + str(f_eval_compact))
        
    # Initialize Morris indices so that the memory is reserved when broadcasting
    if f_eval_compact.ndim == 2:
        morris_mean_abs = np.zeros((morris_samp.shape[1], f_eval_compact.shape[1]), dtype = float) # n_poi x n_qoi
        morris_mean = np.zeros(morris_mean_abs.shape, dtype = float)
        morris_std = np.zeros(morris_mean_abs.shape, dtype = float) # n_poi x n_qoi
    else :
        morris_mean_abs = np.zeros((morris_samp.shape[1]), dtype = float) # n_poi x n_qoi
        morris_mean = np.zeros(morris_mean_abs.shape, dtype = float)
        morris_std = np.zeros(morris_mean_abs.shape, dtype = float) # n_poi x n_qoi
    
    
    # Perform morris calculation only on base thread
    if mpi_rank == 0:
        #Compute # of pois, qois and samples to ensure consitency
        if morris_samp.ndim == 1:
            n_poi = 1
        elif morris_samp.ndim == 2:
            n_poi = morris_samp.shape[1]
        else:
            raise Exception("More than 2 dimensions in morris_samp")
        #Convert to int so it can be used in indexing
        n_samp = int(morris_samp.shape[0]/(n_poi+1))
        if f_eval_compact.ndim == 2:
            n_qoi = f_eval_compact.shape[1]
        elif f_eval_compact.ndim ==1:
            n_qoi = 1
        else:
            raise Exception("More than 2 dimensions in f_eval")
            
        #Uncompact Samples
        f_eval_seperated = morris_seperate(f_eval_compact, n_samp, n_poi, n_qoi)
        morris_samp_seperated = morris_seperate(morris_samp, n_samp, n_poi, n_poi)
        if logging > 1:
            print("morris samp seperated: " + str(morris_samp_seperated))
        
        #Get which sample perturbs which poi
        poi_pert_location = get_poi_pert_location(morris_samp_seperated)
        if logging > 1:
            print("poi_pert_location: " + str(poi_pert_location))
            
            
        #initialize data storage arrays with 1 dimension lower if n_qoi =1
        if n_qoi > 1:
            deriv_approx = np.empty((n_samp, n_poi, n_qoi))  # n_samp x n_poi x n_qoi
        else: 
            deriv_approx = np.empty((n_samp, n_poi))  # n_samp x n_poi
        if logging >1:
            print("QOIs : " + str(f_eval_seperated))
        
        #Apply finite difference formula 
        #Source: Smith, R. 2011, Uncertainty Quanitification. p.333
        if logging > 0:
            print("Calculating Morris indices")
        for i_samp in range(n_samp):
            for i_pert in range(n_poi): 
                i_poi = poi_pert_location[i_samp, i_pert]
                pert = np.sum((morris_samp_seperated[i_samp, i_pert+1] -\
                                       morris_samp_seperated[i_samp, i_pert]))
                #Check exactly one parameter is changed
                if 1 != np.nonzero(morris_samp_seperated[i_samp, i_pert+1] -\
                                       morris_samp_seperated[i_samp, i_pert])[0].size:
                    raise Exception("More than one POI changed in morris stepping.")
                deriv_approx[i_samp,i_poi] = (f_eval_seperated[i_samp,i_pert+1] - \
                                              f_eval_seperated[i_samp,i_pert])/ \
                                              pert
        # for i_poi in range(n_poi):
        #     deriv_approx[:,i_poi] = f_eval_seperated[:,i_poi+1] - f_eval_seperated[:,i_poi]
        if logging > 1:
            print("deriv approx: " + str(deriv_approx))
            
        #Apply Morris Index formulas
        #Source: Smith, R. 2011, Uncertainty Quanitification. p.332
        morris_mean_abs = np.mean(np.abs(deriv_approx),axis = 0) # n_poi x n_qoi
        morris_mean = np.mean(deriv_approx, axis = 0)
        morris_std=np.sqrt(np.var(deriv_approx, axis=0)) # n_poi x n_qoi
        
        if logging > 1:
            print("morris mean abs: " + str(morris_mean_abs))
            print("morris mean abs: " + str(morris_mean))
            print("morris st: " + str(morris_std))
        if logging  and mpi_size > 1: 
            print("Broadcasting Morris Indices")
    #Send out finished morris indices to all threads
    mpi_comm.Bcast([morris_mean_abs, MPI.DOUBLE], root = 0)
    mpi_comm.Bcast([morris_mean, MPI.DOUBLE], root = 0)
    mpi_comm.Bcast([morris_std, MPI.DOUBLE], root = 0)
    if logging:
        print("Finished broadcasting (thread " + str(mpi_rank) + ")")

    return morris_mean_abs, morris_mean, morris_std

def morris_seperate(qoi_compact, n_samp, n_poi, n_qoi):
    if n_qoi > 1:
        qoi_seperated = np.empty((n_samp, n_poi+1, n_qoi))
    else: 
        qoi_seperated = np.empty((n_samp, n_poi+1))
    #Seperate each parameter search for ease of computation
    for i_samp in range(n_samp):
        qoi_seperated[i_samp] = qoi_compact[i_samp*(n_poi+1):(i_samp+1)*(n_poi+1)].squeeze()
        
    return qoi_seperated

def get_poi_pert_location(morris_samp_seperate):
    n_samp = morris_samp_seperate.shape[0]
    n_poi = morris_samp_seperate.shape[2]
    poi_pert_location = np.empty((n_samp, n_poi))
    for i_samp in range(n_samp):
        for i_poi in range(n_poi):
            poi_pert_location[i_samp,i_poi] = np.argmax(np.abs(morris_samp_seperate[i_samp,i_poi+1] \
                                          - morris_samp_seperate[i_samp,i_poi]))
    return poi_pert_location.astype(int)
##---------------------------get_morris_poi_sample-----------------------------

def get_morris_poi_sample(param_dist, n_samp, n_poi, pert_distance, random = False):
    #Use sobol distributions for low discrepancy
    #Generate n_samp_morris samples
    random_samp =  param_dist(n_samp)
    #Define Sampling matrices that are constant
    J=np.ones((n_poi+1,n_poi))
    B = (np.tril(np.ones(J.shape), -1))
    morris_samp_compact = np.empty((n_samp*(n_poi+1), n_poi))
    for i_samp in range(n_samp):
        if callable(pert_distance):
            samp_mat = np.empty(n_poi+1, n_poi)
            samp_mat[0,:] = random_samp[i_samp, :]
            for i_poi in range(n_poi):
                samp_mat[i_poi+1,:] = samp_mat[i_poi,:]
                samp_mat[i_poi+1,i_poi] = pert_distance(samp_mat[i_poi,:], i_poi)
        else: 
            jTheta=random_samp[i_samp,]*J
            #Calculate Morris Sample matrix
            #Source: Smith, R. 2011. Uncertainty Quantification. p.334
            if random == True:  
                #Define Random Sampling matrices
                #D=np.diag(np.random.choice(np.array([1,-1]), size=(n_poi,)))
                #NOTE: using non-random step direction to keep denominator in deriv approx
                #   equal to delta rather than -delta for some samples. Random form is
                #   kept above in comments 
                D=np.diag(np.random.choice(np.array([1,1]), size=(n_poi,)))
                P=np.identity(n_poi)
                np.random.shuffle(P)
                samp_mat = np.matmul(jTheta+pert_distance/2*(np.matmul((2*B-J),D)+J),P)
            elif random == False:
                #Define non-random Sampling matrices
                D=np.diag(np.random.choice(np.array([1,1]), size=(n_poi,)))
                P=np.identity(n_poi)
                np.random.shuffle(P)
                # Only use non-random formulations for testing matrix generation
                samp_mat = jTheta+pert_distance/2*(np.matmul((2*B-J),D)+J)
            #Stack each grid seach so that a single eval_fcn call is required
        morris_samp_compact[i_samp*(n_poi+1):(i_samp+1)*(n_poi+1),:] = samp_mat
    return morris_samp_compact


            




#UQtoolbox
#Module for calculating local sensitivity indices, parameter correlation, and Sobol indices for an arbitrary model
#Authors: Harley Hanes, NCSU, hhanes@ncsu.edu
#Required Modules: numpy, seaborne
#Functions: LSA->GetJacobian
#           GSA->ParamSample, PlotGSA, GetSobol

#3rd party Modules
import numpy as np
import sys
import warnings
import matplotlib.pyplot as plt
import scipy.integrate as integrate
from tabulate import tabulate   

import mpi4py.MPI as MPI

#Package Modules
from . import lsa
from . import gsa
from . import examples
from . import sampling
#import seaborne as seaborne
###----------------------------------------------------------------------------------------------
###-------------------------------------Class Definitions----------------------------------------
###----------------------------------------------------------------------------------------------

##--------------------------------------uqOptions--------------------------------------------------
#Define class "uqOptions", this will be the class used to collect algorithm options for functions
#   -Subclasses: lsaOptions, plotOptions, gsaOptions
#--------------------------------------lsaOptions------------------------------------------------

#--------------------------------------gsaOptions------------------------------------------------

#--------------------------------------plotOptions------------------------------------------------
class PlotOptions:
    __slots__=["run", "n_points", "path"]
    def __init__(self,run=True,n_points=400,path=False):
        self.run=run
        self.n_points=n_points
        self.path=path
        pass
#--------------------------------------uqOptions------------------------------------------------
#   Class holding the above options subclasses
class Options:
    __slots__=["lsa", "plot", "gsa", "display", "save", "path"]
    def __init__(self,lsa=lsa.LsaOptions(),plot=PlotOptions(),gsa=gsa.GsaOptions(), \
                 display=True, save=False, path='..'):
        self.lsa=lsa
        self.plot=plot
        self.gsa=gsa
        self.display=display                       #Whether to print results to terminal
        self.save=save                             #Whether to save results to files
        self.path=path                             #Where to save files
        if self.save and not self.path:
            warnings.warn("Save marked as true but no path given, saving files to current folder.")
            path=''
    pass

##-------------------------------------model------------------------------------------------------
#Define class "model", this will be the class used to collect input information for all functions
class Model:
    __slots__=["base_poi", "base_qoi", "name_poi", "name_qoi", "cov", "eval_fcn", "dist_type",\
               "dist_param", "n_poi", "n_qoi", "sample_fcn"]
    #Model sets should be initialized with base parameter settings, covariance Matrix, and eval function that
    #   takes in a vector of POIs and outputs a vector of QOIs
    def __init__(self,base_poi=np.empty(0), name_poi = "auto", \
                 name_qoi= "auto", cov=np.empty(0), \
                 eval_fcn=np.empty(0), dist_type='uniform', dist_param="auto"):
        #------------------------base_poi, n_poi, name_poi---------------------
        #Assign base_poi and n_poi
        if not isinstance(base_poi,np.ndarray):                    #Confirm that base_poi is a numpy array
            raise Exception("model.base_poi is not a numpy array")
        if np.ndim(base_poi)>1:                                    #Check to see if base_poi is a vector
            base_poi=np.squeeze(base_poi)                     #Make a vector if an array with 1 dim greater than 1
            if np.ndim(base_poi)!=1:                               #Issue an error if base_poi is a matrix or tensor
                raise Exception("Error! More than one dimension of size 1 detected for model.base_poi, model.base_poi must be dimension 1")
            else:                                                       #Issue a warning if dimensions were squeezed out of base POIs
                warnings.warn("model.base_poi was reduced a dimension 1 array. No entries were deleted.")
        self.base_poi=base_poi
        del base_poi
        self.n_poi=self.base_poi.size
        
        #Assign name_poi----------------UNFINISHED VALUE CHECKING
        
        POInumbers=np.arange(0,self.n_poi)
        name_poi_auto=np.char.add('POI',POInumbers.astype('U'))
        #Check name_poi is string
        if type(name_poi)==np.ndarray:
            #Check data type
            if name_poi.size!= self.n_poi:
                raise Exception("Incorrect number of entries in name_poi")
        elif type(name_poi)==list:
            #Check data type
            if len(name_poi)!= self.n_poi:
                raise Exception("Incorrect number of entries in name_poi")
            name_poi=np.array(name_poi)
        elif type(name_poi)==str and name_poi.lower()!="auto":
            if self.n_poi!=1:
                raise Exception("Only one qoi name entered for >1 pois")
        else :
            if name_poi.lower()!= "auto":
                warnings.warn("Unrecognized name_poi entry, using automatic values")
            name_poi = name_poi_auto
        if (name_poi.size != self.n_poi) & (name_poi.size !=0):   #Check that correct size if given
            warnings.warn("name_poi entered but the number of names does not match the number of POIs. Ignoring names.")
            name_poi=np.empty(0)
        if name_poi.size==0:                       
            name_poi = name_poi_auto
        self.name_poi = name_poi
        del name_poi
        #-----------------eval_fcn, base_qoi, n_qoi, name_qoi------------------
        #Assign evaluation function and compute base_qoi
        self.eval_fcn=eval_fcn
        del eval_fcn
        self.base_qoi=self.eval_fcn(self.base_poi)
        if not isinstance(self.base_qoi,np.ndarray):                    #Confirm that base_qoi is a numpy array
            warnings.warn("model.base_qoi is not a numpy array")
        self.n_qoi=len(self.base_qoi)
        
        #Assign name_qoi----------------UNFINISHED VALUE CHECKING
        #Formulate automatic names so they can be referenced in each case
        QOInumbers=np.arange(0,self.n_qoi)
        name_qoi_auto=np.char.add('POI',QOInumbers.astype('U'))
        #Check name_qoi is string
        if type(name_qoi)==np.ndarray:
            #Check data type
            if name_qoi.size!= self.n_qoi:
                warnings.warn("Incorrect number of entries in name_qoi, using automatic")
                name_qoi = "auto"
        elif type(name_qoi)==list:
            #Check data type
            if len(name_qoi)!= self.n_qoi:
                warnings.warn("Incorrect number of entries in name_qoi, using automatic")
                name_qoi = "auto"
            else: 
                name_qoi=np.array(name_qoi)
        elif type(name_qoi)==str and name_qoi.lower()!="auto":
            if self.n_qoi!=1:
                warnings.warn("Incorrect number of entries in name_qoi, using automatic")
                name_qoi = "auto"
            else :
                name_qoi = np.array(name_qoi)
        else :
            if name_qoi.lower()!= "auto":
                warnings.warn("Unrecognized name_qoi entry, using automatic values")
            name_qoi= name_qoi_auto
        self.name_qoi = name_qoi
        del name_qoi
            
            
            
            
        #------------------------------covariance matrix-----------------------
        self.cov=cov
        if self.cov.size!=0 and np.shape(self.cov)!=(self.n_poi,self.n_poi): #Check correct sizing
            raise Exception("Error! model.cov is not an nPOI x nPOI array")
            
        #--------------------------------dist_type-----------------------------
        #Only allow distributions that are currently fully implemented
        valid_distribution =np.array(["uniform", "saltelli uniform", "normal", \
                                      "saltelli normal"])
        # valid_distribution =np.array(["uniform", "normal", "exponential", \
        #                           "saltelli normal", "beta", "InverseCDF"])
        #If distribution type is valid, save its value
        if np.all(dist_type!= valid_distribution):
            raise Exception(str(dist_type) + " is an invalid distribution. Valid" +\
                            " distributions are" + str(valid_distribution))
        else:
            self.dist_type=dist_type
            del dist_type
                
        #--------------------------------dist_param----------------------------
        # Apply automatic distribution parameter settings
        if str(dist_param).lower() == 'auto':
            if (self.dist_type == "uniform" or self.dist_type == "saltelli uniform"):
                self.dist_param=[[.8],[1.2]]*np.ones((2,self.n_poi))*self.base_poi
                del dist_param
            elif (self.dist_type == "normal" or self.dist_type == "saltelli normal"):
                if cov.size()==0:
                    self.dist_param=[[1],[.2]]*np.ones((2,self.n_poi))*self.base_poi
                    del dist_param
        # Apply manual distribution settings
        elif type(dist_param) == np.ndarray:
            #Check dimensions of numpy array are correct
            if dist_param.shape[1] == self.n_poi:
                # Correct number of parameters for each distribution
                if (self.dist_type == "uniform" or self.dist_type == "saltelli uniform")\
                    and dist_param.shape[0]!=2:
                    raise Exception("2 parameters per POI required for uniform")
                elif (self.dist_type == "normal" or self.dist_type == "saltelli normal")\
                    and dist_param.shape[0]!=2:
                    raise Exception("2 parameters per POI required for normal")
                # Assign dist_param if conditions met
                else :
                    self.dist_param=dist_param
                    del dist_param
            else:
                raise Exception("Incorrect shape of dist_param. Given shape: "\
                                + dist_param.shape + ", desired shape: ... x n_poi") 
        elif dist_param.lower() == "cov":
            if np.any(self.dist_type.lower()==["normal", "saltelli normal"]):
                self.dist_param=[self.base_poi, np.diag(self.cov,k=0)]   
                del dist_param 
            else :
                raise Exception("Covariance based sampling only implemented for"\
                                +"normal or saltelli normal distributions")
        else:
            raise Exception("Incorrect data-type for dist_param, use ndarray, 'auto', or 'cov'")
            
        #Construct Distribution function
        self.sample_fcn = sampling.get_samp_dist(self.dist_type, self.dist_param, self.n_poi)
    
    pass
    def copy(self):
        return Model(base_poi=self.base_poi, name_poi = self.name_poi, name_qoi= self.name_qoi, cov=self.cov, \
                 eval_fcn=self.eval_fcn, dist_type=self.dist_type,dist_param=self.dist_param)

##------------------------------------results-----------------------------------------------------
# Define class "results" which holds a gsaResults object and lsaResults object

class Results:
    def __init__(self,lsa=lsa.LsaResults(), gsa=gsa.GsaResults()):
        self.lsa=lsa
        self.gsa=gsa
    pass


###----------------------------------------------------------------------------------------------
###-------------------------------------Main Functions----------------------------------------
###----------------------------------------------------------------------------------------------
#   The following functions are the primary functions for running the package. RunUQ runs both local sensitivity
#   analysis and global sensitivity analysis while printing to command window summary statistics. However, local
#   sensitivity analysis and global sensitivity analysis can be run independently with LSA and GSA respectively

##--------------------------------------RunUQ-----------------------------------------------------
def run_uq(model, options, logging = False):
    """Runs both the local and global sensitivity, and result printing, saving, and plotting.
    
    Parameters
    ----------
    model : Model
        Object of class Model holding run information.
    options : Options
        Object of class Options holding run settings.
    logging : bool or int
        Holds level 
        
    Returns
    -------
    Results 
        Object of class Results holding all run results.
    """
    mpi_comm = MPI.COMM_WORLD
    mpi_rank = mpi_comm.Get_rank()
    mpi_size = mpi_comm.Get_size()
    
    if logging : 
        print("Starting UQ with precessor " + str(mpi_rank) + " of "+ str(mpi_size) + ".")
    
    #Only hold results in base processor
    results = Results()
        
    #Run Local Sensitivity Analysis only on base processor
    if mpi_rank == 0 and options.lsa.run:
        if logging: 
            print("Starting LSA")
        results.lsa = lsa.run_lsa(model, options.lsa, logging = logging)
        #---------------Broadcast results.lsa to other threads-----------

    #Run Global Sensitivity Analysis
    # if options.gsa.run:
        # if options.lsa.run:
        #     #Use a reduced model if it was caluclated
        #     results.gsa=GSA(results.lsa.reducedModel, options)
        # else:
    if options.gsa.run:
        if logging: 
            print("Starting GSA")
        results.gsa = gsa.run_gsa(model, options.gsa, logging = logging)

    #Print, save, and plot Results only on thread 0
    if options.display and mpi_rank ==0:
        if logging: 
            print("Printing Results")
        print_results(results,model,options)                     #Print results to standard output path

    if options.save and mpi_rank == 0:
        if logging: 
            print("Saving Results")
        original_stdout = sys.stdout                            #Save normal output path
        sys.stdout=open(options.path + 'Results.txt', 'a+')            #Change output path to results file
        print_results(results,model,options)                     #Print results to file
        sys.stdout=original_stdout                              #Revert normal output path
        if options.gsa.run_morris:
            np.savez(options.path + "morris_indices.npz",\
                     morris_mean_abs = results.gsa.morris_mean_abs,
                     morris_mean = results.gsa.morris_mean,
                     morris_std = results.gsa.morris_std,
                     base_response = model.base_qoi)

    #Plot Samples
    if options.plot:
        if logging: 
            print("Plotting Results")
        if options.lsa.run_pss and mpi_rank == 0:
            plot_lsa(model, results.lsa.ident_values, options)
        if options.gsa.run_sobol and options.gsa.run  and mpi_rank == 0:
            plot_gsa(model, results.gsa.samp_d, results.gsa.f_d, options)

    return results



def print_results(results,model,options):
    """Prints Results object to console or document.
    
    Parameters
    ----------
    results : Results
        Object of class Model holding run information.
    model : Model
        Object of class Model holding run information.
    options : Options
        Object of class Options holding run settings.
    """
    # Print Results
    #Results Header
    #print('Sensitivity results for nSampSobol=' + str(options.gsa.n_samp_sobol))
    #Local Sensitivity Analysis
    if options.lsa.run:
        print('Local Methods using ' + options.lsa.deriv_method \
              +' approximation and h=' + str(options.lsa.x_delta))
        if options.lsa.run_lsa:
            print('Base POI Values')
            print(tabulate([model.base_poi], headers=model.name_poi))
            print('\n Base QOI Values')
            print(tabulate([model.base_qoi], headers=model.name_qoi))
            print('\n Sensitivity Indices')
            print(tabulate(np.concatenate((model.name_poi.reshape(model.n_poi,1),np.transpose(results.lsa.jac)),1),
                  headers= np.append("",model.name_qoi)))
            print('\n Relative Sensitivity Indices')
            print(tabulate(np.concatenate((model.name_poi.reshape(model.n_poi,1),np.transpose(results.lsa.rsi)),1),
                  headers= np.append("",model.name_qoi)))
            #print("Fisher Matrix: " + str(results.lsa.fisher))
        #Active Subsapce Analysis
        if options.lsa.run_pss:
            print('\nParameter Subset selection using ' + options.lsa.pss_decomp_method +\
                  ' decomposition, ' + str(options.lsa.pss_algorithm) + ' algorithm' + \
                  ' and tolerance ' + str(options.lsa.pss_rel_tol))
            print('Active Supspace')
            print(results.lsa.active_set)
            print('\nInactive Supspace')
            print(results.lsa.inactive_set)
            if options.lsa.pss_algorithm.lower() == "smith":
                print('\nIdentifiability Values')
                for i_sim in range(len(results.lsa.ident_values)):
                    print('Pass ' + str(i_sim+1) + ': ' + str(results.lsa.ident_values[i_sim]))
            elif options.lsa.pss_algorithm.lower() == "rrqr":
                print('\nIdentifiability Values: ' + str(results.lsa.ident_values))
    if options.gsa.run: 
        if options.gsa.run_sobol:
            if model.n_qoi==1:
                print('\n Sobol Indices for ' + model.name_qoi[0])
                print(tabulate(np.concatenate((model.name_poi.reshape(model.n_poi,1), results.gsa.sobol_base.reshape(model.n_poi,1), \
                                               results.gsa.sobol_tot.reshape(model.n_poi,1)), 1),
                               headers=["", "1st Order", "Total Sensitivity"]))
            else:
                for i_qoi in range(0,model.n_qoi):
                    print('\n Sobol Indices for '+ model.name_qoi[i_qoi])
                    print(tabulate(np.concatenate((model.name_poi.reshape(model.n_poi,1),results.gsa.sobol_base[[i_qoi],:].reshape(model.n_poi,1), \
                        results.gsa.sobol_tot[[i_qoi],:].reshape(model.n_poi,1)),1), headers = ["", "1st Order", "Total Sensitivity"]))
    
        if options.gsa.run_morris:
            if model.n_qoi==1:
                print('\n Morris Screening Results for ' + model.name_qoi[0])
                print(tabulate(np.concatenate((model.name_poi.reshape(model.n_poi, 1), results.gsa.morris_mean_abs.reshape(model.n_poi, 1), \
                                               results.gsa.morris_std.reshape(model.n_poi, 1)), 1),
                    headers=["", "mu_star", "sigma"]))
            else:
                for i_qoi in range(model.n_qoi):
                    print('\n Morris Screening Results for ' + model.name_qoi[i_qoi])
                    print(tabulate(np.concatenate(
                        (model.name_poi.reshape(model.n_poi, 1), results.gsa.morris_mean_abs[:,[i_qoi]].reshape(model.n_poi, 1), \
                     results.gsa.morris_std[:,[i_qoi]].reshape(model.n_poi, 1)), 1),
                    headers=["", "mu_star", "sigma"]))

###----------------------------------------------------------------------------------------------
###-------------------------------------Support Functions----------------------------------------
###----------------------------------------------------------------------------------------------
def plot_lsa(model, ident_values, options):
    """Plots Identifiability singular values results from lsa module.
    
    Parameters
    ----------
    model : Model
        Object of class Model holding run information.
    sample_mat: np.ndarray
        n_samp x n_poi array holding each parameter sample
    eval_mat : np.ndarray
        n_samp x n_qoi array holding each function evaluation
    options : Options
        Object of class Options holding run settings.
    """
    fig = plt.figure()
    for i_sim in range(len(ident_values)):
        plt.semilogy(ident_values[i_sim], label = 'Iteration %i' % i_sim)
    fig.tight_layout()
    if options.lsa.pss_decomp_method.lower() == "svd":
        plt.ylabel("Singular Value Magnitude")
        plt.xlabel("Singular Value Number")
    elif options.lsa.pss_decomp_method.lower() == "eigen":
        plt.ylabel("Eigenvalue Magnitude")
        plt.xlabel("Eigenvalue Number")
    plt.xlabel('Identifiability Value ('+ options.lsa.pss_decomp_method + ')')
    plt.legend()
    
    plt.savefig(options.path+"identifiability_values.png")
    

def plot_gsa(model, sample_mat, eval_mat, options):
    """Plots Sobol Sampling results from gsa module.
    
    Parameters
    ----------
    model : Model
        Object of class Model holding run information.
    sample_mat: np.ndarray
        n_samp x n_poi array holding each parameter sample
    eval_mat : np.ndarray
        n_samp x n_qoi array holding each function evaluation
    options : Options
        Object of class Options holding run settings.
    """
    #Reduce Sample number
    #plotPoints=range(0,int(sample_mat.shape[0]), int(sample_mat.shape[0]/plotOptions.n_points))
    #Make the number of sample points to survey
    plotPoints=np.linspace(start=0, stop=sample_mat.shape[0]-1, num=options.plot.n_points, dtype=int)
    #Plot POI-POI correlation and distributions
    figure, axes=plt.subplots(nrows=model.n_poi, ncols= model.n_poi, squeeze=False)
    for iPOI in range(0,model.n_poi):
        for jPOI in range(0,iPOI+1):
            if iPOI==jPOI:
                n, bins, patches = axes[iPOI, jPOI].hist(sample_mat[:,iPOI], bins=41)
            else:
                axes[iPOI, jPOI].plot(sample_mat[plotPoints,iPOI], sample_mat[plotPoints,jPOI],'b*')
            if jPOI==0:
                axes[iPOI,jPOI].set_ylabel(model.name_poi[iPOI])
            if iPOI==model.n_poi-1:
                axes[iPOI,jPOI].set_xlabel(model.name_poi[jPOI])
            if model.n_poi==1:
                axes[iPOI,jPOI].set_ylabel('Instances')
    figure.tight_layout()
    if options.path:
        plt.savefig(options.path+"POIcorrelation.png")

    #Plot QOI-QOI correlationa and distributions
    figure, axes=plt.subplots(nrows=model.n_qoi, ncols= model.n_qoi, squeeze=False)
    for i_qoi in range(0,model.n_qoi):
        for j_qoi in range(0,i_qoi+1):
            if i_qoi==j_qoi:
                axes[i_qoi, j_qoi].hist([eval_mat[:,i_qoi]], bins=41)
            else:
                axes[i_qoi, j_qoi].plot(eval_mat[plotPoints,i_qoi], eval_mat[plotPoints,j_qoi],'b*')
            if j_qoi==0:
                axes[i_qoi,j_qoi].set_ylabel(model.name_qoi[i_qoi])
            if i_qoi==model.n_qoi-1:
                axes[i_qoi,j_qoi].set_xlabel(model.name_qoi[j_qoi])
            if model.n_qoi==1:
                axes[i_qoi,j_qoi].set_ylabel('Instances')
    figure.tight_layout()
    if options.path:
        plt.savefig(options.path+"QOIcorrelation.png")

    #Plot POI-QOI correlation
    figure, axes=plt.subplots(nrows=model.n_qoi, ncols= model.n_poi, squeeze=False)
    for i_qoi in range(0,model.n_qoi):
        for jPOI in range(0, model.n_poi):
            axes[i_qoi, jPOI].plot(sample_mat[plotPoints,jPOI], eval_mat[plotPoints,i_qoi],'b*')
            if jPOI==0:
                axes[i_qoi,jPOI].set_ylabel(model.name_qoi[i_qoi])
            if i_qoi==model.n_qoi-1:
                axes[i_qoi,jPOI].set_xlabel(model.name_poi[jPOI])
    if options.path:
        plt.savefig(options.path+"POI_QOIcorrelation.png")
    #Display all figures
    if options.display:
        plt.show()
        




# -*- coding: utf-8 -*-
"""
Created on Tue Jan 18 14:03:22 2022

@author: USER
"""
#3rd party Modules
import numpy as np
import sys
import scipy as scp
import warnings
#import matplotlib.pyplot as plt
#import scipy.integrate as integrate
#from tabulate import tabulate                       #Used for printing tables to terminal
#import sobol                                        #Used for generating sobol sequences
#import SALib.sample as sample
import scipy.stats as sct
from .sampling import parallel_eval
import mpi4py.MPI as MPI
import matplotlib.pyplot as plt
from matplotlib.ticker import MaxNLocator

class LsaOptions:
    __slots__=["run", "run_lsa", "run_pss", "x_delta", "deriv_method", "scale",\
               "pss_algorithm", "pss_rel_tol", "pss_decomp_method"]
    def __init__(self,run=True, run_lsa = True, run_pss=True, x_delta=10**(-12),\
                 deriv_method='complex', scale='y', pss_algorithm = "rrqr",\
                 pss_rel_tol=1e-8, pss_decomp_method = "svd"):
        #----------------------------Run Selections----------------------------
        self.run=run                              #Whether to run lsa (True or False)
        if self.run == False:
            self.run_pss = False
            self.run_lsa= False
        else:
            self.run_pss=run_pss
            self.run_lsa = run_lsa
        #-------------------------Derivative Calculations----------------------
        self.x_delta=x_delta                      #Input perturbation for calculating jacobian
        self.deriv_method=deriv_method                        #method used for approximating derivatives
        if not self.deriv_method.lower() in ('complex','finite'):
            raise Exception('Error! unrecognized derivative approx method. Use complex or finite')
        if self.x_delta<0 or not isinstance(self.x_delta,float):
            raise Exception('Error! Non-compatibale x_delta, please use a positive floating point number')
        #-----------------------Parameter Subset Selection---------------------
        #pss_rel_tol
        if type(pss_rel_tol)!= float:
            warnings.warn("Invalid data type for pss_rel_tol: " + str(type(pss_rel_tol)) \
                          +". Defaulting to 1e-8")
            self.pss_rel_tol = 1e-8
        elif pss_rel_tol<0 or pss_rel_tol>1 :
            warnings.warn("Invalid pss_rel_tol: " + str(pss_rel_tol)+". Defaulting to 1e-8")
            self.pss_rel_tol = 1e-8
        else :
            self.pss_rel_tol=pss_rel_tol
            
        #pss_decomp_method
        if type(pss_decomp_method)!= str:
            warnings.warn("Non-string value for pss_decomp_method detected. Defaulting to SVD.")
            self.pss_decomp_method = "svd"
        elif pss_decomp_method.lower() != "svd" and pss_decomp_method.lower() != "eigen":
            warnings.warn("Unknown pss_decomp_method: "+str(pss_decomp_method)+". Defaulting to SVD.")
            self.pss_decomp_method = "svd"
        else:
            self.pss_decomp_method = pss_decomp_method
        
        #pss_algorithm
        if type(pss_algorithm)!= str:
            warnings.warn("Non-string value for pss_algorithm detected. Defaulting to RRQR.")
            self.pss_algorithm = "rrqr"
        elif pss_algorithm.lower() != "rrqr" and pss_algorithm.lower() != "rrqr":
            warnings.warn("Unknown pss_algorithm: "+str(pss_algorithm)+". Defaulting to RRQR.")
            self.pss_algorithm = "rrqr"
        else:
            self.pss_algorithm = pss_algorithm
            
    pass


##--------------------------------------LSA-----------------------------------------------------
# Local Sensitivity Analysis main
class LsaResults:
    def __init__(self,jacobian=np.empty, rsi=np.empty, fisher=np.empty,
                 active_set="", inactive_set="", ident_values = np.empty,\
                 ident_vectors=np.empty):
        self.jac=jacobian
        self.rsi=rsi
        self.fisher=fisher
        self.active_set=active_set
        self.inactive_set=inactive_set
        self.ident_values = ident_values
        self.ident_vectors = ident_vectors
    pass


def run_lsa(model, lsa_options, logging = 0):
    """Implements local sensitivity analysis using LSI, RSI, and parameter subset reduction.
    
    Parameters
    ----------
    model : Model
        Object of class Model holding run information.
    options : Options
        Object of class Options holding run settings.
        
    Returns
    -------
    LsaResults 
        Object of class LsaResults holding all run results.
    """
    # LSA implements the following local sensitivity analysis methods on system specified by "model" object
        # 1) Jacobian
        # 2) Scaled Jacobian for Relative Sensitivity Index (RSI)
        # 3) Fisher Information matrix
    # Required Inputs: object of class "model" and object of class "options"
    # Outputs: Object of class lsa with Jacobian, RSI, and Fisher information matrix

    # Calculate Jacobian
    if lsa_options.run_lsa:
        jac_raw=get_jacobian(model.eval_fcn, model.base_poi, lsa_options.x_delta,\
                             lsa_options.deriv_method, scale=False, y_base=model.base_qoi)
        # Calculate relative sensitivity index (RSI)
        jac_rsi=get_jacobian(model.eval_fcn, model.base_poi, lsa_options.x_delta,\
                             lsa_options.deriv_method, scale=True, y_base=model.base_qoi)
        # Calculate Fisher Information Matrix from jacobian
        fisher_mat=np.dot(np.transpose(jac_raw), jac_raw)

    #Active Subspace Analysis
    if lsa_options.run_pss:
        active_set, inactive_set, ident_values, ident_vectors= \
            get_active_subset(model, lsa_options,logging = logging)
        #Collect Outputs and return as an lsa object
    if lsa_options.run_lsa and lsa_options.run_pss:
        return LsaResults(jacobian=jac_raw, rsi=jac_rsi, fisher=fisher_mat,\
                          active_set=active_set, inactive_set=inactive_set,\
                          ident_values = ident_values, ident_vectors = ident_vectors)
    elif lsa_options.run_pss and not lsa_options.run_lsa:
        return LsaResults(active_set=active_set, inactive_set=inactive_set,\
                          ident_values = ident_values, ident_vectors = ident_vectors)
    elif lsa_options.run_lsa and not lsa_options.run_pss:
        return LsaResults(jacobian=jac_raw, rsi=jac_rsi, fisher=fisher_mat)
    
###----------------------------------------------------------------------------------------------
###-------------------------------------Support Functions----------------------------------------
###----------------------------------------------------------------------------------------------

  
    
  
##--------------------------------------GetJacobian-----------------------------------------------------
def get_jacobian(eval_fcn, x_base, x_delta, deriv_method, **kwargs):
    """Calculates scaled or unscaled jacobian using different derivative approximation methods.
    
    Parameters
    ----------
    eval_fcn : 
        Holds run information.
    x_base : np.ndarray
        POI values at which to calculate Jacobian
    lsa_options : Lsa_Options
        Holds run options
    **scale : bool
        Whether or not to apply relative scaling of POI and QOI
    **y_base : np.ndarray
        QOI values used in finite difference approximation, saves a function evaluation
        
    Returns
    -------
    np.ndarray 
        Scaled or unscaled jacobian
    """
    if 'scale' in kwargs:                                                   # Determine whether to scale derivatives
                                                                            #   (for use in relative sensitivity indices)
        scale = kwargs["scale"]
        if not isinstance(scale, bool):                                     # Check scale value is boolean
            raise Exception("Non-boolean value provided for 'scale' ")      # Stop compiling if not
    else:
        scale = False                                                       # Function defaults to no scaling
    if 'y_base' in kwargs:
        y_base = eval_fcn(x_base)
        #y_base = kwargs["y_base"]
        # Make sure x_base is int/ float and convert to numpy array
        if type(y_base)==int or type(y_base)==float:
            y_base = np.array([y_base])
        elif type(x_base)== list:
            y_list = y_base
            y_base = np.empty(len(y_list))
            for i_poi in len(y_list):
                if type(y_list[i_poi])==int or type(y_list[i_poi])==float:
                   y_base[i_poi] = y_list[i_poi]
                else:
                    raise Exception(str(i_poi) + "th y_base value is of type:  " + str(type(y_list[i_poi])))
        elif type(y_base)!= np.ndarray:
            raise Exception("y_base of type " + str(type(y_base)) + ". Accepted" \
                            " types are int/ float and list or numpy arrays of ints/ floats")
    else:
        y_base = eval_fcn(x_base)

    # Make sure x_base is int/ float and convert to numpy array
    if type(x_base)==int or type(x_base)==float:
        x_base = np.array([x_base])
    elif type(x_base)== list:
        x_list = x_base
        x_base = np.empty(len(x_list))
        for i_poi in len(x_list):
            if type(x_list[i_poi])==int or type(x_list[i_poi])==float:
               x_base[i_poi] = x_list[i_poi]
            else:
                raise Exception(str(i_poi) + "th x_base value is of type:  " + str(type(x_list[i_poi])))
    elif type(x_base)!= np.ndarray:
        raise Exception("x_base of type " + str(type(x_base)) + ". Accepted" \
                        " types are int/ float and list or numpy arrays of ints/ floats")

    #Initialize base QOI value, the number of POIs, and number of QOIs
    n_poi = np.size(x_base)
    n_qoi = np.size(y_base)

    jac = np.empty(shape=(n_qoi, n_poi), dtype=float)                       # Define Empty Jacobian Matrix

    for i_poi in range(0, n_poi):                                            # Loop through POIs
        # Isolate Parameters
        if deriv_method.lower()== 'complex':
            xPert = x_base + np.zeros(shape=x_base.shape)*1j                  # Initialize Complex Perturbed input value
            xPert[i_poi] += x_delta * 1j                                      # Add complex Step in input
        elif deriv_method.lower() == 'finite':
            xPert = x_base.copy()
            xPert[i_poi] += x_delta
        yPert = eval_fcn(xPert)                                        # Calculate perturbed output
        for i_qoi in range(0, n_qoi):                                        # Loop through QOIs
            if deriv_method.lower()== 'complex':
                jac[i_qoi, i_poi] = np.imag(yPert[i_qoi] / x_delta)                 # Estimate Derivative w/ 2nd order complex
            elif deriv_method.lower() == 'finite':
                jac[i_qoi, i_poi] = (yPert[i_qoi]-y_base[i_qoi]) / x_delta
            #Only Scale Jacobian if 'scale' value is passed True in function call
            if scale:
                jac[i_qoi, i_poi] *= x_base[i_poi] * np.sign(y_base[i_qoi]) / (sys.float_info.epsilon + y_base[i_qoi])
                                                                            # Scale jacobian for relative sensitivity
        del xPert, yPert, i_poi, i_qoi                                        # Clear intermediate variables
    return jac                                                              # Return Jacobian




##--------------------------------------------Parameter dimension reduction------------------------------------------------------

def get_active_subset(model, lsa_options,logging=0):
    """ Selects active subset algorithm and returns results.
    Parameters
    ----------
    model : Model
        Holds run information.
    lsa_options : Lsa_Options
        Holds run options
        
    Returns
    -------
    Model 
        New model using reduced parameters
    np.ndarray
        Data type string of active parameters
    np.ndarray
        Data type string of inactive parameters
    """
    
    if lsa_options.pss_algorithm.lower()=="smith":
        if logging:
            print("Running Smith Textbook Subset Selection Algoirthm")
        active_set, inactive_set, ident_values, ident_vectors= \
            pss_smith(model.eval_fcn, model.base_poi, model.base_qoi, \
                              model.name_poi, model. name_qoi, lsa_options.pss_decomp_method,\
                              lsa_options.pss_rel_tol, lsa_options.x_delta, \
                              lsa_options.deriv_method, logging = logging)
    elif lsa_options.pss_algorithm.lower() == "rrqr":
        if logging:
            print("Running RRQR Parameter Subset Selection Algoirthm")
        active_set, inactive_set, ident_values, ident_vectors = \
            pss_rrqr(model.eval_fcn, model.base_poi, model.base_qoi, \
                              model.name_poi, model. name_qoi, lsa_options.pss_decomp_method,\
                              lsa_options.pss_rel_tol, lsa_options.x_delta, \
                              lsa_options.deriv_method, logging = logging)
    else :
        raise Exception("Unrecognized pss_algorith: " + str(lsa_options.pss_algorith))
    return active_set, inactive_set, ident_values, ident_vectors
        
def pss_rrqr(eval_fcn, base_poi, base_qoi, name_poi, name_qoi,\
                      decomp_method, subset_rel_tol, x_delta, deriv_method,
                      logging =0):
    """Calculates active and inactive parameter subsets according to algorithm B1 of
        [requires reference]. Algorithm outline is as follows.
        1) Compute Sensitivity Matrix
        2) Compute svd of sensitivity matrix or eigendecomposition of Fisher
        3) Identify unidentifiable parameters
        4) Check error of reduced model
    
    Parameters
    ----------
    model : Model
        Holds run information.
    lsa_options : Lsa_Options
        Holds run options
        
    Returns
    -------
    Model 
        New model using reduced parameters
    np.ndarray
        Data type string of active parameters
    np.ndarray
        Data type string of inactive parameters
    """
    n_poi = base_poi.size
    unidentifiable_indexes = np.array([], dtype = int)
    identifiable_indexes = np.arange(0,n_poi,dtype = int)
    #Step 1) Calculate Sensitivity matrix
    sens = get_jacobian(eval_fcn, base_poi, x_delta,\
                         deriv_method, scale=False, y_base=base_qoi)
    #Step 2) Compute eigen/ singular value decomposition
    #Perform Eigendecomp
    if decomp_method.lower() == "eigen":
        #Caclulate Fisher
        fisher_mat=np.dot(np.transpose(sens), sens)
        ident_values, ident_vectors =np.linalg.eig(fisher_mat)
    elif decomp_method.lower() == "svd":
        #Perform QR Decomposition to improve efficiency
        Q, R =np.linalg.qr(sens, mode='reduced')
        u, ident_values, ident_vectors = np.linalg.svd(R, full_matrices = False)
    if logging >1 : 
        print("Initial Identifiability Values: " + str(ident_values))
        print("Initial Identifiability Vectors: "  + str(ident_vectors))
    #Step 3) Identify unidentifiable parameters
    unidentifiable_singular_values = np.nonzero((ident_values/ident_values[0])<subset_rel_tol)
    num_unidentifiable = np.size(unidentifiable_singular_values)
    num_identifiable = n_poi - num_unidentifiable
    if logging > 1:
        print("Num unidentifiable parameter: " + str(num_unidentifiable))
    if num_unidentifiable >0:
        #Get identifiability vector
        unidentifiable_vec = ident_vectors[:,-1]
        max_ind = np.argmax(np.abs(unidentifiable_vec))
        
        #Move largest magnitude element to bottom of vector
        P_tild = np.arange(0,n_poi)
        P_tild[[max_ind, -1]] = P_tild[[-1, max_ind]]
        
        #Compute QR of R*P_tild
        B = R[:, P_tild]
        Q_tild, R_tild = np.linalg.qr(B)
        
        Q = np.matmul(Q,Q_tild)
        R = R_tild
        P = P_tild
        if logging > 2:
            print("Q shape: "+ str(Q.shape))
            print("R: "+ str(R))
        del max_ind 
        for iteration in range(num_unidentifiable-1):
            #l is max index of remaining parameters to be reduced, so if n_poi =4
            # and n_unidentifiable = 2, then l=2
            l = n_poi - iteration-1
            R_11 = R[0:l,0:l]
            R_12 = R[0:l, (1+l):]
            R_22 = R[(1+l):, (1+l):]
            null1, null2, V_l = np.linalg.svd(R_11)
            v_l = V_l[:,-1]
            
            max_ind = np.argmax(np.abs(v_l))
            P_tild_loop = np.arange(0,v_l.shape[0])
            P_tild_loop[[-1, max_ind]] = P_tild_loop[[max_ind, -1]]
            
            R_11P = R_11[:, P_tild_loop]
            Q_tild_smlr, R_11_tild = np.linalg.qr(R_11P)
            if logging > 2:
                print("R_11P: "+ str(R_11P))
                print("Q_tild_smlr: "+ str(Q_tild_smlr))
                print(scp.linalg.block_diag(Q_tild_smlr, np.eye(n_poi-l)))
            Q = np.matmul(Q, scp.linalg.block_diag(Q_tild_smlr, np.eye(n_poi-l)))
            P[0:l] = P[P_tild_loop]
            R = scp.linalg.block_diag(R_11_tild, R_22)
            R[0:l, (l+1):] = np.matmul(np.transpose(Q_tild_smlr),R_12)
        identifiable_indexes = P[0:num_identifiable]
        unidentifiable_indexes = P[num_identifiable:]
    #Define active and inactive spaces
    active_set=name_poi[identifiable_indexes]
    inactive_set=name_poi[unidentifiable_indexes]
    
    # reduced_model = model_reduction(model, inactive_param)
    # reduced_model.base_poi=reduced_model.base_poi[inactive_index == False]
    # reduced_model.name_poi=reduced_model.name_poi[inactive_index == False]
    # reduced_model.eval_fcn = lambda reduced_poi: model.eval_fcn(
    #     np.array([x for x, y in zip(reduced_poi,model.base_poi) if inactive_index== True]))
    # #reduced_model.eval_fcn=lambda reduced_poi: model.eval_fcn(np.where(inactive_index==False, reduced_poi, model.base_poi))
    # reduced_model.base_qoi=reduced_model.eval_fcn(reduced_model.base_poi)
    return active_set, inactive_set, ident_values, ident_vectors

def pss_smith(eval_fcn, base_poi, base_qoi, name_poi, name_qoi,\
                      decomp_method, subset_rel_tol, x_delta, deriv_method,
                      logging =0):
    """Calculates active and inactive parameter subsets.
        --Not fully function, reduced model is still full model
    
    Parameters
    ----------
    model : Model
        Holds run information.
    lsa_options : Lsa_Options
        Holds run options
        
    Returns
    -------
    Model 
        New model using reduced parameters
    np.ndarray
        Data type string of active parameters
    np.ndarray
        Data type string of inactive parameters
    """
    eliminate=True
    n_poi = base_poi.size
    inactive_index=np.zeros(n_poi)
    #Calculate Jacobian
    jac=get_jacobian(eval_fcn, base_poi, x_delta,\
                         deriv_method, scale=False, y_base=base_qoi)
    #Inititalize lists that hold the values and vectors at each iteration
    ident_values_stored = []
    ident_vectors_stored = []
    reduction_order = []
    unidentifiable_poi = 0
    while eliminate:
        #Perform Eigendecomp
        if decomp_method.lower() == "eigen":
            #Caclulate Fisher
            fisher_mat=np.dot(np.transpose(jac), jac)
            ident_values, ident_vectors =np.linalg.eig(fisher_mat)
        elif decomp_method.lower() == "svd":
            u, ident_values, ident_vectors = np.linalg.svd(jac, full_matrices = False)
            # if logging >1 : 
            #     print("Identifiability Values: " + str(ident_values))
            #     print("Identifiability Vectors: "  + str(ident_vectors))
            
        ident_values_stored.append(ident_values)
        ident_vectors_stored.append(ident_vectors)
            
        
        #Eliminate dimension/ terminate
        if ident_values[-unidentifiable_poi-1] < subset_rel_tol * np.max(ident_values):
            unidentifiable_poi+=1
            #Get inactive parameter
            inactive_param = np.argmax(np.absolute(ident_vectors[:, -unidentifiable_poi]))
                #This indexing may seem odd but its because we're keeping the full model parameter numbering while trying
                # to index within the reduced model so we have to add to the index the previously removed params
            #Record inactive param in inactive space
            if logging > 1 :
                print("Unidentifiable Param Index: " + str(inactive_param))
                print("Corresponding Vector: " + str(ident_vectors[:, -unidentifiable_poi]))
                # if np.all(ident_vectors[:, np.argmin(np.absolute(ident_values))] == ident_vectors[:,-1]):
                #     print("Unidentifiable vector is last vector")
                # else :
                #     print("Unidentifiable vector is NOT last vector")
            inactive_index[inactive_param]=1
            reduction_order.append(name_poi[inactive_param])
            #Zero out inactive elements of jacobian
            jac[:,inactive_param] = 0
        else:
            #Terminate Active Subspace if singular values within tolerance
            eliminate=False
            
    #Define active and inactive spaces
    active_set=name_poi[inactive_index == False]
    inactive_set=name_poi[inactive_index == True]

# reduced_model = model_reduction(model, inactive_param)
# reduced_model.base_poi=reduced_model.base_poi[inactive_index == False]
# reduced_model.name_poi=reduced_model.name_poi[inactive_index == False]
# reduced_model.eval_fcn = lambda reduced_poi: model.eval_fcn(
#     np.array([x for x, y in zip(reduced_poi,model.base_poi) if inactive_index== True]))
# #reduced_model.eval_fcn=lambda reduced_poi: model.eval_fcn(np.where(inactive_index==False, reduced_poi, model.base_poi))
# reduced_model.base_qoi=reduced_model.eval_fcn(reduced_model.base_poi)
    return active_set, inactive_set, ident_values_stored, ident_vectors_stored


def test_model_reduction(model, inactive_pois, n_samp, save_location,pss_tol, \
                         logging = 0):
    mpi_comm = MPI.COMM_WORLD
    mpi_rank = mpi_comm.Get_rank()
    mpi_size = mpi_comm.Get_size()
    # Get inactive indices
    inactive_indices = np.where(np.isin(model.name_poi, inactive_pois))
    if logging:
        print("Inactive Indices:" + str(inactive_indices))
        print("Inactive POIs:" + str(model.name_poi[inactive_indices]))
    # Generate Parameter Sample
    poi_samp_full = model.sample_fcn(n_samp)
    poi_samp_reduced = np.copy(poi_samp_full)
    # Fix Inactive Parameters
    poi_samp_reduced[:,inactive_indices] = model.base_poi[inactive_indices]

    print("poi_samp_full:" +str(poi_samp_full))
    print("poi_samp_reduced:" +str(poi_samp_reduced))
    # Compute model evaluations
    if mpi_size == 0: 
        qoi_samp_full = model.eval_fcn(poi_samp_full)
        qoi_samp_reduced = model.eval_fcn(poi_samp_reduced)
    else : 
        qoi_samp_full = parallel_eval(model.eval_fcn, poi_samp_full, logging = logging)
        qoi_samp_reduced = parallel_eval(model.eval_fcn, poi_samp_reduced, logging =logging)
    mpi_comm.Barrier()
    
    print("qoi_samp_full:" +str(qoi_samp_full))
    print("qoi_samp_reduced:" +str(qoi_samp_reduced))
    if mpi_rank==0:
        #Save results
        np.savez(save_location + "data.npz",\
                 qoi_samp_full = qoi_samp_full, qoi_samp_reduced = qoi_samp_reduced,\
                 poi_samp_reduced = poi_samp_reduced, poi_samp_full = poi_samp_full)
        for i_qoi in range(model.n_qoi):
            
            # Compute KDE approximations
            kernel_full = sct.gaussian_kde(qoi_samp_full[:,i_qoi])
            kernel_reduced = sct.gaussian_kde(qoi_samp_reduced[:,i_qoi])
            
            #Plot KDEs
            plot_kde(kernel_full, kernel_reduced, \
                     [np.min(qoi_samp_full[:,i_qoi]), np.max(qoi_samp_full[:,i_qoi])],\
                     model.name_qoi[i_qoi], 
                     save_location + "kde_"+ model.name_qoi[i_qoi] + ".png",\
                     pss_tol)
                
def test_model_reduction_precomputed(name_qoi, qoi_samp_full, qoi_samp_reduced, n_samp, save_location,pss_tol, \
                         logging = 0, fontsize = 14, linewidth = 2.5):
    mpi_comm = MPI.COMM_WORLD
    mpi_rank = mpi_comm.Get_rank()
    mpi_size = mpi_comm.Get_size()
    if mpi_rank==0:
        for i_qoi in range(name_qoi.size):
            
            # Compute KDE approximations
            kernel_full = sct.gaussian_kde(qoi_samp_full[:,i_qoi])
            kernel_reduced = sct.gaussian_kde(qoi_samp_reduced[:,i_qoi])
            
            #Plot KDEs
            plot_kde(kernel_full, kernel_reduced, \
                     [np.min(qoi_samp_full[:,i_qoi]), np.max(qoi_samp_full[:,i_qoi])],\
                     name_qoi[i_qoi], 
                     save_location + "kde_"+  name_qoi[i_qoi] + ".png",\
                     pss_tol,
                     fontsize = fontsize,
                     linewidth = linewidth)


        
def plot_kde(kernel_full, kernel_reduced, qoi_bounds, qoi_name, save_location, pss_tol,
             fontsize = 14, linewidth = 2.5):
    #Generate qoi evaluation values
    qoi_spacing = np.linspace(qoi_bounds[0], qoi_bounds[1],200)
    #Evaluate kernels at qoi values
    qois_full = kernel_full.pdf(qoi_spacing)
    qois_reduced = kernel_reduced.pdf(qoi_spacing)
    #Plot kernels
    fig = plt.figure()
    plt.rc('font', size= fontsize) 
    plt.plot(qoi_spacing, qois_full, linewidth = linewidth ,label="Full Model")
    plt.plot(qoi_spacing, qois_reduced, linewidth = linewidth, label="Reduced Model", linestyle='dashdot')
    plt.xlabel(qoi_name)
    plt.title("KDE of " + qoi_name +  " (PSS tolerance=" + str(pss_tol) + ")")
    plt.gca().xaxis.set_major_locator(MaxNLocator(nbins = 4))
    plt.gca().yaxis.set_major_locator(MaxNLocator(nbins = 4))
    plt.legend()
    #fig.tight_layout()
    plt.savefig(save_location)
    


def model_reduction(model,inactive_param):
    """Computes a new Model object using only active parameter set"
        -Not fully function, reduced model is still full model
        
    Parameters
    ----------
    model : Model
        Original run information
    inactive_param : Lsa_Options
        Holds run options
    
        
    Returns
    -------
    Model 
        New model using reduced parameters
    """
    reduced_model = model
    # #Record Index of reduced param
    # inactive_index=np.where(reduced_model.name_poi==inactive_param)[0]
    # #confirm exactly parameter matches
    # if len(inactive_index)!=1:
    #     raise Exception("More than one or no POIs were found matching that name.")
    # #Remove relevant data elements
    # reduced_model.base_poi=np.delete(reduced_model.base_poi, inactive_index)
    # reduced_model.name_poi=np.delete(reduced_model.name_poi, inactive_index)
    # reduced_model.eval_fcn=lambda reduced_poi: model.eval_fcn(np.where(inactive_index==True,reduced_poi,model.base_poi))
    # print('made eval_fcn')
    # print(reduced_model.eval_fcn(reduced_model.base_poi))
    return reduced_model

def get_reduced_pois(reduced_poi,dropped_indices,model):
    """Maps from the space of reduced pois to the full poi set.
        
    Parameters
    ----------
    reduced_poi : np.ndarray
        Set of reduced parameter set values
    dropped_indices : np.ndarray
        Indices of dropped parameters
    model : Model
        Original run information
    
        
    Returns
    -------
    np.ndarray 
        New model using reduced parameters
    """
    #Load in full parameter set to start output
    full_poi=model.base_poi
    #Use a counter to keep track of the index in full_pois based on index in
    # reduced_pois
    reduced_counter=0
    for i_poi in np.arange(0,model.n_poi):
        if dropped_indices==i_poi:
            full_poi[i_poi]=reduced_poi[reduced_counter]
            reduced_counter=reduced_counter+1
            
            
    return full_poi


# -*- coding: utf-8 -*-
"""
Created on Wed Mar 16 20:25:03 2022

@author: Harley Hanes
"""

from setuptools import setup, find_packages

from pathlib import Path
this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text(encoding="utf-8")

setup(
    name='UQLibrary',
    version='0.1.3',
    author='H. Hanes',
    author_email='hhanes@ncsu.edu',
    packages=find_packages(include=["UQLibrary", "UQLibrary.*"]),
    url='https://github.com/HarleyHanes/UQLibrary-CourseProject',
    license='LICENSE',
    description='Robust set of sensitivity and identifiability analysis methods.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    install_requires=[
        "numpy >= 1.20.0",
        "scipy >= 1.7.1",
        "matplotlib >= 3.4.3",
        "tabulate >= 0.8.9",
        "mpi4py >= 3.1.3"
    ],
    python_requires='>=3.8',
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)


# -*- coding: utf-8 -*-
"""
Created on Tue Mar 29 19:26:38 2022

@author: USER
"""
import numpy as np
from scipy.stats import qmc
import scipy.stats as sct

import mpi4py.MPI as MPI

#======================================================================================================
#-----------------------------------Sampling-----------------------------------------------------------
#======================================================================================================
       

##--------------------------------------GetSampDist----------------------------------------------------
def get_samp_dist(dist_type, dist_param, n_poi, fcn_inverse_cdf = np.nan):
    """Adds sampling function sample to model for drawing of low-discrepency
        from given distribution type.
    
    Parameters
    ----------
    model : Model
        Contaings simulation information.
    gsa_options : GSAOptions
        Contains run settings
        
    Returns
    -------
    Model
        model object with added sample function
    """
    # Determine Sample Function- Currently only 1 distribution type can be defined for all parameters
    if dist_type == 'normal':  # Normal Distribution
        sample_fcn = lambda n_samp_sobol: np.random.randn(n_samp_sobol, n_poi)*\
            np.sqrt(dist_param[[1], :]) + dist_param[[0], :]
    elif dist_type == 'saltelli normal':
        sample_fcn = lambda n_samp_sobol: saltelli_normal(n_samp_sobol, dist_param)
    elif dist_type == 'uniform':  # uniform distribution
        # doubleParms=np.concatenate(model.dist_param, model.dist_param, axis=1)
        sample_fcn = lambda n_samp_sobol: np.random.rand(n_samp_sobol, n_poi)*\
            (dist_param[[1], :]-dist_param[[0],:]) + dist_param[[0], :]
    elif dist_type == 'saltelli uniform':  # uniform distribution
        # doubleParms=np.concatenate(model.dist_param, model.dist_param, axis=1)
        sample_fcn = lambda n_samp_sobol: saltelli_uniform(n_samp_sobol, dist_param)
    elif dist_type == 'exponential': # exponential distribution
        sample_fcn = lambda n_samp_sobol: np.random.exponential(dist_param,size=(n_samp_sobol, n_poi))
    elif dist_type == 'beta': # beta distribution
        sample_fcn = lambda n_samp_sobol:np.random.beta(dist_param[[0],:], dist_param[[1],:],\
                                               size=(n_samp_sobol, n_poi))
    elif dist_type == 'InverseCDF': #Arbitrary distribution given by inverse cdf
        if fcn_inverse_cdf == np.nan:
            raise Exception("InverseCDF distribution selected but no function provided.")
        sample_fcn = lambda n_samp_sobol: fcn_inverse_cdf(np.random.rand(n_samp_sobol, n_poi))
    else:
        raise Exception("Invalid value for model.dist_type. Supported distributions are normal, uniform, exponential, beta, \
        and InverseCDF")  # Raise Exception if invalide distribution is entered
    
    return sample_fcn



def saltelli_sample(n_samp, n_poi):
    """Constructs a uniform [0,1] low discrepency saltelli sample for use in Sobol
        index approximation
    
    Parameters
    ----------
    n_samp_sobol : int
        Number of samples to take
    dist_param : np.ndarray
        2 x n_poi array of min and max sampling bounds for each parameter
        
    Returns
    -------
    np.ndarray
        Low discrepancy POI sample of uniform distribution on [0,1] constructed 
        using satelli's alrogrithm
    """
    
    #Add .5 to n_samp/2 so that if n_samp is odd, an extra sample is generated
    sampler = qmc.Sobol(d= n_poi*2, scramble = True)
    #Use the smallest log2 sample size at least as large as n_samp to keep
    #   quadrature balance 
    #   (see https://scipy.github.io/devdocs/reference/generated/scipy.stats.qmc.Sobol.html )
    base_sample = sampler.random_base2(m=int(np.ceil(np.log2(n_samp/2))))
    
    #Add .5 to n_samp/2 so that if n_samp is odd, an extra sample is generated
    base_sample=base_sample[:int(n_samp/2+.5),:]
    
    sample = np.empty((n_samp, n_poi))
    
    #Seperate and stack half the samples in the 2nd dimension for saltelli's 
    # algorithm
    if n_samp%2==0:
        sample[:int((n_samp)/2),:]=base_sample[:,0:n_poi]
        sample[int((n_samp)/2):,:]=base_sample[:,n_poi:]
    else :
        sample[:int((n_samp+.5)/2),:] = base_sample[:,0:n_poi]
        sample[int((n_samp+.5)/2):-1,:] = base_sample[:,n_poi:]
    return sample


def saltelli_uniform(n_samp, dist_param):
    """Constructs a uniform low discrepency saltelli sample for use in Sobol
        index approximation
    
    Parameters
    ----------
    n_samp: int
        Number of samples to take
    dist_param : np.ndarray
        2 x n_poi array of mean and variance for each parameter
        
    Returns
    -------
    np.ndarray
        Low discrepancy POI sample of uniform distribution constructed using 
        satelli's alrogrithm
    """
    n_poi=dist_param.shape[1]
    
    sample_base = saltelli_sample(n_samp,n_poi)
    
    sample_transformed = dist_param[[0],:]+(dist_param[[1],:]-dist_param[[0],:])*sample_base
    return sample_transformed


def saltelli_normal(n_samp, dist_param):
    """Constructs a normal low discrepency saltelli sample for use in Sobol
        index approximation
    
    Parameters
    ----------
    n_samp: int
        Number of samples to take
    dist_param : np.ndarray
        2 x n_poi array of mean and variance for each parameter
        
    Returns
    -------
    np.ndarray
        Low discrepancy POI sample of normal distribution constructed using 
        satelli's alrogrithm
    """
    
    n_poi=dist_param.shape[1]
    
    sample_base = saltelli_sample(n_samp,n_poi)
    sample_transform=sct.norm.ppf(sample_base)*np.sqrt(dist_param[[1], :]) \
        + dist_param[[0], :]
    return sample_transform


#======================================================================================================
#----------------------------Parallelization Support---------------------------------------------------
#======================================================================================================

def parallel_eval(eval_fcn, poi_sample, logging = False):
    """ Seperates samples and parallelizes model computations

    Parameters
    ----------
    eval_fcn : function
        User defined function that maps POIs to QOIs
    poi_sample : np.ndarray
        n_samp x n_poi array of POI samples

    Returns
    -------
    qoi_samp : np.ndarray
        n_samp x n_qoi array of QOIs from each POI sample
    """
    mpi_comm = MPI.COMM_WORLD
    mpi_rank = mpi_comm.Get_rank()
    mpi_size = mpi_comm.Get_size()
     
    # Seperate poi samples into subsample for each thread
    if mpi_rank == 0:
        if logging > 1:
            print("poi_sample in thread " + str(mpi_rank) + ": " + str(poi_sample))
        for i_rank in range(mpi_size):
            if mpi_rank == 0:
                samp_per_subsample = int(np.round(poi_sample.shape[0]/mpi_size))
                if i_rank == 0:
                    data = poi_sample[0:samp_per_subsample]
                else: 
                    if i_rank == mpi_size-1:
                        data_broadcast = poi_sample[(i_rank*samp_per_subsample):]
                    else:
                        data_broadcast = poi_sample[(i_rank*samp_per_subsample):((i_rank+1)*samp_per_subsample)]
                    mpi_comm.send(data_broadcast.shape, dest = i_rank, tag = 0)
                    mpi_comm.Send([data_broadcast,MPI.DOUBLE],dest = i_rank, tag = 1)
                    #print("poi_subsample sent to thread " + str(i_rank) + ": " + str(data_broadcast))
    else:
        data_shape = mpi_comm.recv(source = 0, tag = 0)
        data = np.empty(data_shape)
        mpi_comm.Recv(data,source=0, tag=1)
                
    
    # Evaluate each subsamples
    qoi_subsample = eval_fcn(data)
    mpi_comm.Barrier()
    if qoi_subsample.ndim == 2:
        qoi_sample = np.zeros((poi_sample.shape[0], qoi_subsample.shape[1]), dtype = float)
    else:
        qoi_sample = np.zeros((poi_sample.shape[0]), dtype = float)
    #print(poi_reconstructed)

    if mpi_rank > 0:
        mpi_comm.send(qoi_subsample.shape, dest = 0, tag = 0)
        mpi_comm.Send([qoi_subsample, MPI.DOUBLE], dest = 0, tag =1)
        #print("sending data from thread " + str(mpi_rank) + ": " + str(data))
    elif mpi_rank ==0 :
        total_samp=0
        for i_rank in range(mpi_size):
            if i_rank > 0:
                subsample_shape = mpi_comm.recv(source = i_rank, tag = 0)
                #print("receiving data from thread " + str(i_rank) + " of shape: " + str(data_shape))
            else :
                subsample_shape = qoi_subsample.shape
            n_samp = subsample_shape[0]
            if i_rank > 0:
                #print("poi_reconstructed before receiving: " + str(poi_reconstructed))
                mpi_comm.Recv(qoi_sample[total_samp:(total_samp+n_samp)], source = i_rank, tag=1)
            else :
                qoi_sample[total_samp:(total_samp+n_samp)] = qoi_subsample
            if logging > 1:
                print("qoi_reconstructed after receiving thread " + str(i_rank) + ": " + str(qoi_sample))
            total_samp += n_samp 
            
    # Send back out qoi_sample so all threads have a return
    mpi_comm.Bcast([qoi_sample, MPI.DOUBLE], root = 0)
    
    
    return qoi_sample
